import threading


class Curses:
    __singleton_lock = threading.Lock()
    __singleton_instance = None

    @classmethod
    def __new__(cls, *args, **kwargs):
        if not cls.__singleton_instance:
            with cls.__singleton_lock:
                if not cls.__singleton_instance:
                    cls.__singleton_instance = super().__new__(cls)
        return cls.__singleton_instance

    def __init__(self):
        temp = __import__('curses')
        self._screen = temp.initscr()
        for key in temp.__dict__:
            if key != 'initscr' and not key.startswith('_'):
                setattr(self, key, getattr(temp, key, None))

    def initscr(self):
        return self._screen


curses = Curses()



import os
import threading

from .meta import BareException
from .editor import Editor
from .const import CHAR_DICT, HELP_GUIDE
from .console import curses


class Window:
    __singleton_lock = threading.Lock()
    __singleton_instance = None

    @classmethod
    def __new__(cls, *args, **kwargs):
        if not cls.__singleton_instance:
            with cls.__singleton_lock:
                if not cls.__singleton_instance:
                    cls.__singleton_instance = super().__new__(cls)
        return cls.__singleton_instance

    def __init__(self, app):
        self.app = app
        self.config = app.config
        self.colors = {}
        self.c = ''

        # Initialize curses window
        self.screen = curses.initscr()

        curses.noecho()  # turns off echoing of keys to the screen
        curses.cbreak()  # read keys without requiring Enter key

        self.screen.keypad(1)  # Enable keypad mode (read special keys)

        self.height, self.width = self.screen.getmaxyx()  # GRABS SIZE of CURRENT TERMINAL WINDOW
        # (self.height, self.width) = windowSize  # split into height and width

        self.height -= 1
        self.width -= 1

        self._win = curses.newwin(self.height, self.width, 0, 0)  # 0,0 is start position
        self.editor = Editor(self)

    def print_header(self, save_info='', message='', save_path='', total=0):
        """Prints header to curses screen"""
        if message:
            # Print info message
            self.screen.addstr(0, 0, (' ' * self.width),
                               self.config['color_header'])  # this was commented out, not sure why
            self.screen.addstr(0, 0, message, self.config['color_message'])
        else:
            self.screen.addstr(0, 0, (' ' * self.width), self.config['color_header'])  # print empty line
            if not save_path:
                self.screen.addstr(0, 0, 'file/not/saved', self.config['color_header'])  # print file/not/saved
            elif len(save_path) > self.width - 14:
                temp_text = os.path.split(save_path)[1]
                self.screen.addstr(0, 0, '%s%s' % (temp_text, save_info),
                                   self.config['color_header'])  # print filename only
            else:
                self.screen.addstr(0, 0, '%s%s' % (save_path, save_info),
                                   self.config['color_header'])  # print directory and filename

        temp_text = '%i' % total
        lines_text = temp_text.rjust(11)
        if self.config['inline_commands'] == 'protected':
            protect_string = str(self.config['protect_string'])
            self.screen.addstr(0, self.width - 12 - len(protect_string) - 1, lines_text, self.config['color_header'])
            self.screen.addstr(0, self.width - len(protect_string) - 1, protect_string, self.config['color_message'])
        else:
            self.screen.addstr(0, self.width - 12, lines_text, self.config['color_header'])

        self.screen.hline(1, 0, curses.ACS_HLINE, self.width, self.config['color_bar'])

    def addstr(self, *args, **kwargs):
        self.screen.addstr(*args, **kwargs)

    def hline(self, *args, **kwargs):
        self.screen.hline(*args, **kwargs)

    def getch(self, *args, **kwargs):
        self.c = self.screen.getch(*args, **kwargs)
        if self.c == 10 and self.config.debug:
            self.screen.nodelay(1)
            while self.screen.getch() >= 0:
                pass
            self.screen.nodelay(0)

        return self.c

    def vline(self, *args, **kwargs):
        self.screen.vline(*args, **kwargs)

    def refresh(self, *args, **kwargs):
        self.screen.refresh(*args, **kwargs)

    def clear(self, *args, **kwargs):
        self.screen.clear(*args, **kwargs)

    def keypad(self, *args, **kwargs):
        self.screen.keypad(*args, **kwargs)

    def status_message(self, text, number, total, update_lines=False):
        """Displays status message"""
        if update_lines:  # clears entire header and updates number of lines
            self.addstr(0, 0, ' ' * self.width, self.config['color_header'])
            temp_text = f'{total:d}'
            lines_text = temp_text.rjust(11)
            if self.config['inline_commands'] == 'protected':
                protect_string = str(self.config['protect_string'])
                self.addstr(0, self.width - 12 - len(protect_string) - 1, lines_text, self.config['color_header'])
                self.addstr(0, self.width - len(protect_string) - 1, protect_string, self.config['color_message'])
            else:
                self.addstr(0, self.width - 12, lines_text, self.config['color_header'])
        else:  # clears space for statusMessage only
            self.addstr(0, 0, " " * (self.width - 13), self.config['color_header'])
        number = int(number)  # Convert to integer
        message = f' {text}{number:d}% '
        self.addstr(0, 0, message, self.config['color_warning'])
        self.refresh()

    def get_confirmation(self, text=' Are you sure? (y/n) ', any_key=False, x=0, y=0):
        """Confirm selection in new (centered) window. Returns 'True' if 'y' pressed."""
        if not any_key and self.config['skip_confirmation'] and text != ' File exists, overwrite? (y/n) ':
            return True
        side = '   '
        if len(text) < 15:
            diff = 15 - len(text)
            spacer = ' ' * int(diff / 2)
            text = spacer + text + spacer
        line = (len(text) + (len(side) * 2)) * ' '

        half_height = int(self.height / 2)
        self.hline(half_height - 1, int(self.width / 2) - int(len(text) / 2) - len(side),
                   curses.ACS_HLINE, (len(text) + 6), self.config["color_message"])
        # print corners
        self.hline(half_height - 1, int(self.width / 2) - int(len(text) / 2) - len(side),
                   curses.ACS_ULCORNER, 1, self.config["color_message"])
        self.hline(half_height - 1, int(self.width / 2) - int(len(text) / 2) + len(text) + 2,
                   curses.ACS_URCORNER, 1, self.config["color_message"])

        self.addstr(half_height + 1, int(self.width / 2) - int(len(text) / 2) - len(side),
                    line, self.config["color_message"])  # Prints blank line
        self.addstr(half_height, int(self.width / 2) - int(len(text) / 2) - len(side),
                    side, self.config["color_message"])  # Prints left side
        self.vline(half_height, int(self.width / 2) - int(len(text) / 2) - len(side),
                   curses.ACS_VLINE, 3, self.config["color_message"])  # prints left side

        self.addstr(half_height, int(self.width / 2) - int(len(text) / 2),
                    text, self.config["color_message"])  # Prints text message

        self.addstr(half_height, int(self.width / 2) - int(len(text) / 2) + len(text),
                    side, self.config["color_message"])  # Prints right side
        self.vline(half_height, int(self.width / 2) - int(len(text) / 2) + len(text) + 2,
                   curses.ACS_VLINE, 3, self.config["color_message"])  # prints right side

        if any_key:
            self.addstr(half_height + 1, int(self.width / 2) - int(len("(press any key)") / 2),
                        "(press any key)", self.config["color_message"])
        self.hline(half_height + 2, int(self.width / 2) - int(len(text) / 2) - len(side),
                   curses.ACS_HLINE, (len(text) + 6), self.config["color_message"])
        # print bottom corners
        self.hline(half_height + 2, int(self.width / 2) - int(len(text) / 2) - len(side),
                   curses.ACS_LLCORNER, 1, self.config["color_message"])
        self.hline(half_height + 2, int(self.width / 2) - int(len(text) / 2) + len(text) + 2,
                   curses.ACS_LRCORNER, 1, self.config["color_message"])

        # Prints text message
        try:
            self.addstr(y + self.height - 2, x, "",
                        self.config["color_normal"])  # Moves cursor to previous position
        except BareException:
            pass

        self.refresh()
        self.c = ''
        while True:
            self.c = self.getch()
            if any_key and self.c:
                return True
            if self.c == ord('y') or self.c == ord('Y'):
                pos = text.find("y")
                self.addstr(half_height, (int(self.width / 2) - int(len(text) / 2) + pos), 'y',
                            self.config["color_quote_double"])  # Prints text message
                self.refresh()
                return True
            if self.c == ord('n') or self.c == ord('N'):
                return False

    def prompt_user(self, title='ENTER COMMAND:', default_answer='',
                    footer="(press 'enter' to proceed, UP arrow to cancel)", adjust_pos=False):
        """
        Displays window and prompts user to enter command
        Used for 'Entry', 'Find', and 'Save' windows
        """

        self.c = ''
        if adjust_pos and '.' in default_answer and default_answer.rfind('.') == len(default_answer) - 4:
            position = len(default_answer) - 4
        elif adjust_pos and '.' in default_answer and default_answer.rfind('.') == len(default_answer) - 3:
            position = len(default_answer) - 3
        else:
            position = len(default_answer)
        text = default_answer
        # side = '   '
        line = int(self.width - 16) * ' '
        if self.width < 70 and footer == "(press 'enter' to proceed, UP arrow to cancel)":
            footer = '(press UP arrow to cancel)'
        footer = footer.center(int(self.width - 16) - 6)
        empty_line = (int(self.width - 16) - 6) * ' '
        if len(text) > len(empty_line) * 2:
            text = text[0:(len(empty_line) * 2)]

        half_height = int(self.height / 2)
        while self.c != 10:
            for i in range(0, 6):
                self.addstr(half_height - 2 + i, 8, line, self.config['color_message'])
            self.addstr(half_height - 1, 11, title, self.config['color_message'])
            self.addstr(half_height + 2, 11, footer, self.config['color_message'])
            self.addstr(half_height, 11, empty_line, self.config['color_normal'])
            self.addstr(half_height + 1, 11, empty_line, self.config['color_normal'])

            # print border
            self.hline(half_height - 2, 9, curses.ACS_HLINE, (len(empty_line) + 4), self.config['color_message'])
            self.hline(half_height + 3, 9, curses.ACS_HLINE, (len(empty_line) + 4), self.config['color_message'])
            self.vline(half_height - 2, 8, curses.ACS_VLINE, 6, self.config['color_message'])
            self.vline(half_height - 2, (len(empty_line) + 13), curses.ACS_VLINE, 6, self.config['color_message'])
            self.hline(half_height - 2, 8, curses.ACS_ULCORNER, 1, self.config['color_message'])
            self.hline(half_height + 3, 8, curses.ACS_LLCORNER, 1, self.config['color_message'])
            self.hline(half_height - 2, (len(empty_line) + 13), curses.ACS_URCORNER, 1, self.config['color_message'])
            self.hline(half_height + 3, (len(empty_line) + 13), curses.ACS_LRCORNER, 1, self.config['color_message'])

            if len(text) > len(empty_line):
                self.addstr(half_height, 11, text[0:len(empty_line)], self.config['color_normal'])
                self.addstr(half_height + 1, 11, text[len(empty_line):], self.config['color_normal'])
            else:
                self.addstr(half_height, 11, text, self.config['color_normal'])

                if len(text) == len(empty_line) and position == len(empty_line):
                    self.addstr(half_height + 1, 11, '', self.config['color_normal'])  # Moves cursor to second line

            # Move cursor
            if position < len(empty_line):
                self.addstr(half_height, position + 11, '', self.config['color_normal'])
            else:
                self.addstr(half_height + 1, position + 11 - len(empty_line), '', self.config['color_normal'])

            self.refresh()
            c = self.getch()

            part1 = text[0:position]
            part2 = text[position:]

            if c in (curses.KEY_UP, self.config['key_find'],
                     self.config['key_entry_window'], self.config['key_save_as']):
                return False
            if c == curses.KEY_LEFT:
                position -= 1
            if c == curses.KEY_RIGHT:
                position += 1
            position = max(0, min(position, len(text)))

            if c == curses.KEY_BACKSPACE or c == 127:
                try:
                    text = part1[0:-1] + part2
                except BareException:
                    pass
                position -= 1

            elif c in CHAR_DICT:
                text = (part1 + CHAR_DICT[c] + part2)
                position += 1
        # attempt to hide encrypt password during load/save
        self.addstr(half_height, 11, empty_line, self.config['color_normal'])
        self.addstr(half_height + 1, 11, empty_line, self.config['color_normal'])
        self.refresh()

        return text

    def print_background(self):
        """Displays background color"""
        for i in range(0, self.height + 2):
            try:
                self.addstr(i, 0, (' ' * self.width), self.config['color_background'])
            except BareException:
                return

    def draw_line_number_background(self):
        """Draws background for line numbers"""
        for y in range(2, self.height + 1):
            self.addstr(y, 0, '     ', self.config['color_line_numbers'])  # Prints blank line number block

    def draw_page_guide(self, end_pos=0):  # , hline_pos=1):
        """Draws page guide"""
        end_pos = end_pos or self.height + 1
        if self.width <= (self.config['page_guide'] + 6):
            return

        for i in range(2, end_pos):
            self.vline(i, (self.config['page_guide'] + 6), curses.ACS_VLINE, 1,
                       self.config['color_bar'])  # prints vertical line
        self.hline(1, (self.config['page_guide'] + 6), curses.ACS_TTEE, 1, self.config['color_bar'])

    def color_on(self, default_colors=False):
        """Turn on curses color and handle color assignments"""
        # global program_message
        self.editor.reset_line()

        if curses.has_colors():
            curses.start_color()
        else:
            if self.config.os_name == 'Macintosh':
                self.editor.get_confirmation('Color not supported on the OSX terminal!', True)
            else:
                self.editor.get_confirmation('Color not supported on your terminal!', True)
            self.config.set_default_settings(True, True)
            self.config['display_color'] = False
            self.editor.program_message = ' Monochrome display '
            return

        curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLACK)
        curses.init_pair(2, curses.COLOR_CYAN, curses.COLOR_BLACK)
        curses.init_pair(3, curses.COLOR_BLUE, curses.COLOR_BLACK)
        curses.init_pair(4, curses.COLOR_GREEN, curses.COLOR_BLACK)
        curses.init_pair(5, curses.COLOR_YELLOW, curses.COLOR_BLACK)
        curses.init_pair(6, curses.COLOR_RED, curses.COLOR_BLACK)
        curses.init_pair(7, curses.COLOR_MAGENTA, curses.COLOR_BLACK)

        curses.init_pair(8, curses.COLOR_BLACK, curses.COLOR_WHITE)
        curses.init_pair(9, curses.COLOR_BLACK, curses.COLOR_CYAN)
        curses.init_pair(10, curses.COLOR_BLACK, curses.COLOR_BLUE)
        curses.init_pair(11, curses.COLOR_BLACK, curses.COLOR_GREEN)
        curses.init_pair(12, curses.COLOR_BLACK, curses.COLOR_YELLOW)
        curses.init_pair(13, curses.COLOR_BLACK, curses.COLOR_RED)
        curses.init_pair(14, curses.COLOR_BLACK, curses.COLOR_MAGENTA)

        curses.init_pair(15, curses.COLOR_BLUE, curses.COLOR_WHITE)
        curses.init_pair(16, curses.COLOR_GREEN, curses.COLOR_WHITE)
        curses.init_pair(17, curses.COLOR_RED, curses.COLOR_WHITE)

        curses.init_pair(18, curses.COLOR_WHITE, curses.COLOR_BLUE)
        curses.init_pair(19, curses.COLOR_WHITE, curses.COLOR_GREEN)
        curses.init_pair(20, curses.COLOR_WHITE, curses.COLOR_RED)

        curses.init_pair(21, curses.COLOR_RED, curses.COLOR_BLUE)
        curses.init_pair(22, curses.COLOR_BLUE, curses.COLOR_RED)

        curses.init_pair(23, curses.COLOR_MAGENTA, curses.COLOR_GREEN)
        curses.init_pair(24, curses.COLOR_GREEN, curses.COLOR_MAGENTA)

        curses.init_pair(25, curses.COLOR_YELLOW, curses.COLOR_GREEN)
        curses.init_pair(26, curses.COLOR_GREEN, curses.COLOR_YELLOW)

        curses.init_pair(27, curses.COLOR_WHITE, curses.COLOR_YELLOW)
        curses.init_pair(28, curses.COLOR_WHITE, curses.COLOR_MAGENTA)
        curses.init_pair(29, curses.COLOR_YELLOW, curses.COLOR_BLUE)
        curses.init_pair(30, curses.COLOR_GREEN, curses.COLOR_BLUE)
        curses.init_pair(31, curses.COLOR_MAGENTA, curses.COLOR_BLUE)
        curses.init_pair(32, curses.COLOR_CYAN, curses.COLOR_BLUE)

        curses.init_pair(33, curses.COLOR_CYAN, curses.COLOR_WHITE)
        curses.init_pair(34, curses.COLOR_YELLOW, curses.COLOR_WHITE)
        curses.init_pair(35, curses.COLOR_MAGENTA, curses.COLOR_WHITE)
        curses.init_pair(36, curses.COLOR_WHITE, curses.COLOR_WHITE)
        curses.init_pair(37, curses.COLOR_YELLOW, curses.COLOR_YELLOW)
        curses.init_pair(38, curses.COLOR_BLACK, curses.COLOR_BLACK)
        curses.init_pair(39, curses.COLOR_GREEN, curses.COLOR_RED)
        curses.init_pair(40, curses.COLOR_YELLOW, curses.COLOR_RED)
        curses.init_pair(41, curses.COLOR_CYAN, curses.COLOR_RED)
        curses.init_pair(42, curses.COLOR_MAGENTA, curses.COLOR_RED)
        curses.init_pair(43, curses.COLOR_BLUE, curses.COLOR_GREEN)
        curses.init_pair(44, curses.COLOR_CYAN, curses.COLOR_GREEN)
        curses.init_pair(45, curses.COLOR_RED, curses.COLOR_GREEN)
        curses.init_pair(46, curses.COLOR_RED, curses.COLOR_YELLOW)
        curses.init_pair(47, curses.COLOR_WHITE, curses.COLOR_CYAN)
        curses.init_pair(48, curses.COLOR_BLUE, curses.COLOR_CYAN)
        curses.init_pair(49, curses.COLOR_RED, curses.COLOR_CYAN)
        curses.init_pair(50, curses.COLOR_YELLOW, curses.COLOR_CYAN)
        curses.init_pair(51, curses.COLOR_MAGENTA, curses.COLOR_CYAN)

        curses.init_pair(52, curses.COLOR_BLUE, curses.COLOR_YELLOW)

        self.colors.update({
            "white_on_black": curses.color_pair(1),
            "cyan_on_black": curses.color_pair(2),
            "blue_on_black": curses.color_pair(3),
            "green_on_black": curses.color_pair(4),
            "yellow_on_black": curses.color_pair(5),
            "red_on_black": curses.color_pair(6),
            "magenta_on_black": curses.color_pair(7),

            "black_on_white": curses.color_pair(8),
            "black_on_cyan": curses.color_pair(9),
            "black_on_blue": curses.color_pair(10),
            "black_on_green": curses.color_pair(11),
            "black_on_yellow": curses.color_pair(12),
            "black_on_red": curses.color_pair(13),
            "black_on_magenta": curses.color_pair(14),

            "blue_on_white": curses.color_pair(15),
            "green_on_white": curses.color_pair(16),
            "red_on_white": curses.color_pair(17),

            "white_on_blue": curses.color_pair(18),
            "white_on_green": curses.color_pair(19),
            "white_on_red": curses.color_pair(20),

            "red_on_blue": curses.color_pair(21),
            "blue_on_red": curses.color_pair(22),

            "magenta_on_green": curses.color_pair(23),
            "green_on_magenta": curses.color_pair(24),

            "yellow_on_green": curses.color_pair(25),
            "green_on_yellow": curses.color_pair(26),

            "white_on_yellow": curses.color_pair(27),
            "white_on_magenta": curses.color_pair(28),

            "yellow_on_blue": curses.color_pair(29),
            "green_on_blue": curses.color_pair(30),
            "magenta_on_blue": curses.color_pair(31),
            "cyan_on_blue": curses.color_pair(32),

            "cyan_on_white": curses.color_pair(33),
            "yellow_on_white": curses.color_pair(34),
            "magenta_on_white": curses.color_pair(35),
            "white_on_white": curses.color_pair(36),
            "yellow_on_yellow": curses.color_pair(37),
            "black_on_black": curses.color_pair(38),
            "green_on_red": curses.color_pair(39),
            "yellow_on_red": curses.color_pair(40),
            "cyan_on_red": curses.color_pair(41),
            "magenta_on_red": curses.color_pair(42),
            "blue_on_green": curses.color_pair(43),
            "cyan_on_green": curses.color_pair(44),
            "red_on_green": curses.color_pair(45),
            "red_on_yellow": curses.color_pair(46),
            "white_on_cyan": curses.color_pair(47),
            "blue_on_cyan": curses.color_pair(48),
            "red_on_cyan": curses.color_pair(49),
            "yellow_on_cyan": curses.color_pair(50),
            "magenta_on_cyan": curses.color_pair(51),

            "blue_on_yellow": curses.color_pair(52),
        })

        if self.config.no_bold:
            bold = 0
        else:
            bold = curses.A_BOLD
        underline = curses.A_UNDERLINE

        # default colors

        if default_colors or self.config["default_colors"]:
            self.config.settings.update({
                "color_dim": self.colors["white_on_black"],
                "color_line_numbers": self.colors["black_on_yellow"],
                "color_line_num_reversed": self.colors["white_on_blue"] + bold,
                "color_warning": self.colors["white_on_red"] + bold,
                "color_normal": self.colors["white_on_black"] + bold,
                "color_background": self.colors["white_on_black"] + bold,
                "color_message": self.colors["white_on_magenta"] + bold,
                "color_reversed": self.colors["white_on_magenta"] + bold,
                "color_underline": self.colors["white_on_black"] + underline + bold,
                "color_commands": self.colors["green_on_black"] + bold,
                "color_commands_reversed": self.colors["white_on_green"] + bold,
                "color_quote_double": self.colors["yellow_on_black"] + bold,
                "color_comment": self.colors["yellow_on_black"],
                "color_comment_block": self.colors["black_on_yellow"],
                "color_comment_separator": self.colors["black_on_red"],
                "color_comment_leftjust": self.colors["white_on_magenta"] + bold,
                "color_comment_rightjust": self.colors["white_on_red"] + bold,
                "color_comment_centered": self.colors["yellow_on_green"] + bold,
                "color_number": self.colors["cyan_on_black"],
                "color_entry": self.colors["white_on_blue"] + bold,

                "color_entry_command": self.colors["green_on_blue"] + bold,
                "color_entry_quote": self.colors["yellow_on_blue"] + bold,
                "color_entry_quote_triple": self.colors["red_on_blue"] + bold,
                "color_entry_comment": self.colors["red_on_blue"] + bold,
                "color_entry_functions": self.colors["magenta_on_blue"] + bold,
                "color_entry_class": self.colors["cyan_on_blue"] + bold,
                "color_entry_number": self.colors["cyan_on_blue"] + bold,
                "color_entry_dim": self.colors["white_on_blue"],

                "color_operator": self.colors["white_on_black"],
                "color_functions": self.colors["magenta_on_black"] + bold,
                "color_functions_reversed": self.colors["white_on_magenta"] + bold,
                "color_class": self.colors["blue_on_black"] + bold,
                "color_class_reversed": self.colors["white_on_blue"] + bold,
                "color_quote_triple": self.colors["red_on_black"],
                "color_mark": self.colors["yellow_on_blue"] + bold + underline,
                "color_negative": self.colors["red_on_black"] + bold,
                "color_entry_negative": self.colors["red_on_blue"] + bold,
                "color_positive": self.colors["cyan_on_black"] + bold,
                "color_entry_positive": self.colors["cyan_on_blue"] + bold,
                "color_tab_odd": self.colors["white_on_black"],
                "color_tab_even": self.colors["yellow_on_black"],
                "color_whitespace": self.colors["black_on_white"] + underline,
                "color_header": self.colors["white_on_black"] + bold,
                "color_bar": self.colors["white_on_black"],
                "color_constant": self.colors["white_on_black"] + underline,
                "color_entry_constant": self.colors["white_on_blue"] + bold,
                "color_quote_single": self.colors["yellow_on_black"] + bold,
                "color_selection": self.colors["black_on_white"] + underline,
                "color_selection_reversed": self.colors["black_on_cyan"] + underline,
            })
        self.config["display_color"] = True

    def curses_off(self):
        """Turns off curses and resets terminal to normal"""
        curses.nocbreak()
        self.keypad(0)
        curses.echo()  # to turn off curses settings
        curses.endwin()  # restore terminal to original condition

    def print_formatted_text(self, y, string, _type='', width=79):
        """Formats curses text by looking for 'special' characters.

            Type can be "rjust" for right justification, "center" for centered.
            Width should be passed when using Type.

            Text formatting
            ---------------
            '_' = UNDERLINE
            '^' = BOLD
            '*' = REVERSE

            String Replacement
            ------------------
            '$' = DIAMOND
            '|" = Vertical Line"""

        underline = False
        bold = False
        reverse = False
        temp_string = string.replace('*', '')  # REVERSE
        temp_string = temp_string.replace('_', '')  # UNDERLINE
        temp_string = temp_string.replace('^', '')  # BOLD

        if _type == 'rjust':
            x = width - len(temp_string)
        elif _type == 'center':
            x = int((width - len(temp_string)) / 2)
        else:
            x = 0

        for z in range(0, len(string)):  # easy way to make first letter of each word standout
            item = string[z]
            if item == '_':
                underline = True
            elif item == '^':
                bold = True
            elif item == '*':
                reverse = True
            elif item == '$':
                self.hline(y, x, curses.ACS_DIAMOND, 1, self.config['color_normal'])  # print diamond
                x += 1
            elif item == '|' and reverse:
                self.vline(y, x, curses.ACS_VLINE, 1, self.config['color_reversed'])  # prints vertical line
                reverse = False
                x += 1
            elif item == '|' and bold:
                self.vline(y, x, curses.ACS_VLINE, 1, self.config.bold_text)  # prints vertical line
                reverse = False
                x += 1
            elif item == '|':
                self.vline(y, x, curses.ACS_VLINE, 1, self.config['color_bar'])  # prints vertical line
                self.hline(y - 1, x, curses.ACS_TTEE, 1, self.config['color_bar'])  # Format previous line
                underline = False
                x += 1
            elif underline:
                underline = False
                self.addstr(y, x, item, self.config['color_underline'])
                x += 1
            elif bold:
                self.addstr(y, x, item, self.config.bold_text)
                bold = False
                x += 1
            elif reverse:
                self.addstr(y, x, item, self.config['color_reversed'])
                reverse = False
                x += 1
            else:
                self.addstr(y, x, item, self.config['color_header'])
                x += 1

    def set_colors(self):
        """Function that allows user to set colors used with syntax highlighting"""
        # global program_message
        self.editor.reset_line()
        if not self.config['display_color'] or not curses.has_colors():
            self.editor.get_confirmation("You can't set colors in monochrome mode!", True)
            return
        if self.width < 79 or self.height < 19:
            self.editor.get_confirmation('Increase terminal size to set colors!', True)
            return

        self.config['default_colors'] = False
        # win = \
        curses.newwin(self.height, self.width, 0, 0)  # 0,0 is start position
        x = int((self.width - 49) / 2)
        c = 0
        i_num = 0
        item_list = []
        c_num = 0
        color_list = []
        temp_list = []
        empty = ''.center(49)
        separator = ''.center(49, '@')
        style = 0
        # style_change = False

        for key in self.config.settings.keys():
            if 'color_' in key:
                item_list.append(key)
        for key in self.colors.keys():
            (item1, item2) = key.split('_on_')
            temp_list.append((item2 + item1, key))  # change "white_on_blue" to ("bluewhite", "white_on_blue")
        temp_list.sort()

        for value in temp_list:
            color_list.append(value[1])

        item_list.sort()
        color_list.insert(0, '[CURRENT]')

        for i in range(0, self.height + 1):
            self.addstr(i, 0, (' ' * self.width), self.config['color_normal'])
            if i <= 8:
                self.addstr(i, x, empty, curses.A_NORMAL)  # redundant?
        title = 'SETCOLORS'.center(49)
        header = ' ITEM (up/down)               COLOR (left/right)'
        # divider = "".center(49, '-')
        # footer = '*N*o*r*m*a*l $ _Bold $ _Underline $ b_Oth'

        sample_header = 'SAMPLE LAYOUT'.center(49)
        sample_left = 'Left justified'.ljust(49)
        sample_right = 'Right justified'.rjust(49)

        while c != 10:  # continue until 'enter' is pressed
            item = item_list[i_num]
            color = color_list[c_num]
            self.addstr(1, x, title, curses.A_REVERSE)
            self.addstr(2, x, header, curses.A_BOLD)
            self.hline(3, x, curses.ACS_HLINE, 49, self.config['color_bar'])
            if color == '[CURRENT]':
                search = ''
                for key, value in self.colors.items():
                    if self.config[item] == value:
                        search = key
                        style = 0
                    elif self.config[item] == value + curses.A_BOLD:
                        search = key
                        style = curses.A_BOLD
                    elif self.config[item] == value + curses.A_UNDERLINE:
                        search = key
                        style = curses.A_UNDERLINE
                    elif self.config[item] == value + curses.A_BOLD + curses.A_UNDERLINE:
                        search = key
                        style = curses.A_BOLD + curses.A_UNDERLINE
                index = color_list.index(search, 1)
                c_num = index
                color = color_list[c_num]

            self.addstr(4, x + 23, (color.replace('_', ' ').rjust(25)), self.colors[color] + style)  # testing
            self.addstr(4, x + 1, (item.replace('color', "").replace('_', ' ')).ljust(23),
                        self.colors['white_on_blue'] + curses.A_BOLD)  # testing
            self.hline(5, x, curses.ACS_HLINE, 49, self.config['color_bar'])
            # print vertical lines
            self.hline(4, x, curses.ACS_VLINE, 1, self.config['color_bar'])
            self.hline(4, x + 48, curses.ACS_VLINE, 1, self.config['color_bar'])
            # print corners
            self.hline(3, x, curses.ACS_ULCORNER, 1, self.config['color_bar'])
            self.hline(3, x + 48, curses.ACS_URCORNER, 1, self.config['color_bar'])
            self.hline(5, x, curses.ACS_LLCORNER, 1, self.config['color_bar'])
            self.hline(5, x + 48, curses.ACS_LRCORNER, 1, self.config['color_bar'])

            if style == curses.A_BOLD + curses.A_UNDERLINE:
                footer = '_Normal $ _Bold $ _Underline $ *b*O*t*h'
            elif style == curses.A_BOLD:
                footer = '_Normal $ *B*o*l*d $ _Underline $ b_Oth'
            elif style == curses.A_UNDERLINE:
                footer = '_Normal $ _Bold $ *U*n*d*e*r*l*i*n*e $ b_Oth'
            else:
                footer = '*N*o*r*m*a*l $ _Bold $ _Underline $ b_Oth'
            self.print_formatted_text(6, footer, 'center', self.width)
            self.addstr(8, x, sample_header, self.config['color_comment_centered'])  # Text types need to be changed?
            self.addstr(9, x, separator, self.config['color_comment_separator'])
            self.addstr(10, x, sample_left, self.config['color_comment_leftjust'])
            self.addstr(11, x, sample_right, self.config['color_comment_rightjust'])

            self.addstr(12, x, 'class', self.config['color_class'])
            self.addstr(12, x + 12, 'collapsed', self.config['color_class_reversed'])
            self.addstr(12, x + 28, 'print', self.config['color_commands'])
            self.addstr(12, x + 40, '#comment', self.config['color_comment'])

            self.addstr(13, x, 'def', self.config['color_functions'])
            self.addstr(13, x + 12, 'collapsed', self.config['color_functions_reversed'])
            self.addstr(13, x + 28, 'True', self.config['color_positive'])
            self.addstr(13, x + 40, 'False', self.config['color_negative'])

            self.addstr(14, x, "'quote'", self.config['color_quote_single'])
            self.addstr(14, x + 12, '"double"', self.config['color_quote_double'])
            self.addstr(14, x + 28, '"""doc"""', self.config['color_quote_triple'])

            self.addstr(14, x + 40, 'CONSTANT', self.config["color_constant"])

            self.addstr(15, x, '()!=[]+-', self.config['color_operator'])
            self.addstr(15, x + 12, 'normal text', self.config['color_normal'])
            self.addstr(15, x + 28, '0123456789', self.config['color_number'])
            self.addstr(15, x + 40, ' C.BLOCK', self.config['color_comment_block'])

            self.addstr(16, x, 'print ', self.config['color_entry_command'])
            self.addstr(16, x + 6, '"Entry line"', self.config['color_entry_quote'])
            self.addstr(16, x + 18, '; ', self.config['color_entry_dim'])
            self.addstr(16, x + 20, 'number ', self.config['color_entry'])
            self.addstr(16, x + 27, '= ', self.config['color_entry_dim'])
            self.addstr(16, x + 29, '100', self.config['color_entry_number'])
            self.addstr(16, x + 32, '; ', self.config['color_entry_dim'])
            self.addstr(16, x + 34, 'def ', self.config['color_entry_functions'])
            self.addstr(16, x + 38, '#comment  ', self.config['color_entry_comment'])

            self.addstr(17, x, 'class', self.config['color_entry_class'])
            self.addstr(17, x + 5, ': ', self.config['color_entry_dim'])
            self.addstr(17, x + 7, 'False', self.config['color_entry_negative'])
            self.addstr(17, x + 12, ', ', self.config['color_entry_dim'])
            self.addstr(17, x + 14, 'True', self.config['color_entry_positive'])
            self.addstr(17, x + 18, '; ', self.config['color_entry_dim'])
            self.addstr(17, x + 20, 'CONSTANT', self.config['color_entry_constant'])
            self.addstr(17, x + 28, '; ', self.config['color_entry_dim'])
            self.addstr(17, x + 30, '"""Triple Quote"""', self.config['color_entry_quote_triple'])

            self.addstr(18, x, '999 ', self.config['color_line_numbers'])
            self.addstr(18, x + 6, '....', self.config['color_tab_odd'])
            self.addstr(18, x + 10, '....', self.config['color_tab_even'])
            self.addstr(18, x + 14, '                                  ', self.config['color_background'])

            self.addstr(19, x + 12, 'Press [RETURN] when done!', self.config['color_warning'])
            self.refresh()
            c = self.getch()

            if c == curses.KEY_UP:
                i_num -= 1
                if i_num < 0:
                    i_num = 0
                c_num = 0
                # style_change = False
            elif c == curses.KEY_DOWN:
                i_num += 1
                if i_num > len(item_list) - 1:
                    i_num = len(item_list) - 1
                c_num = 0
                # style_change = False
            elif c == curses.KEY_LEFT:
                c_num -= 1
                if c_num < 1:
                    c_num = 1
                # style_change = False
                self.config[item_list[i_num]] = self.colors[color_list[c_num]] + style

            elif c == curses.KEY_RIGHT:
                c_num += 1
                if c_num > len(color_list) - 1:
                    c_num = len(color_list) - 1
                # style_change = False
                self.config[item_list[i_num]] = self.colors[color_list[c_num]] + style
            elif c in (ord('b'), ord('B')):
                style = curses.A_BOLD
                # style_change = True
                self.config[item_list[i_num]] = self.colors[color_list[c_num]] + style
            elif c in (ord('u'), ord('U')):
                style = curses.A_UNDERLINE
                # style_change = True
                self.config[item_list[i_num]] = self.colors[color_list[c_num]] + style
            elif c in (ord('n'), ord('N')):  # set style to normal
                style = 0
                # style_change = True  # no longer needed?
                self.config[item_list[i_num]] = self.colors[color_list[c_num]] + style
            elif c in (ord('o'), ord('O')):
                style = curses.A_BOLD + curses.A_UNDERLINE
                # style_change = True
                self.config[item_list[i_num]] = self.colors[color_list[c_num]] + style

    def show_help(self):
        """Display help guide"""
        # global HELP_GUIDE, current_num, saved_since_edit
        # over_sized = False

        try:
            if self.editor.lines.db:
                del self.editor.lines.db
                self.editor.lines.db = {}
        except BareException:
            pass
        self.editor.current_num = 0
        total_rows = 0
        for i in range(0, len(HELP_GUIDE)):
            text = HELP_GUIDE[i]

            line = self.editor.lines.add(text)

            total_rows += (line.number_of_rows - 1)
            if line.number <= (self.height - 2) and self.editor.current_num + total_rows < (self.height - 2):
                self.editor.current_num += 1

        self.editor.current_num -= 1
        self.config.copy_settings()
        self.config['debug'] = False
        self.config['show_indent'] = False
        self.config['entry_highlighting'] = False
        self.config['syntax_highlighting'] = True
        self.config['format_comments'] = True
        self.config['live_syntax'] = True
        self.config['showSpaces'] = False
        self.config['splitscreen'] = False
        self.editor.lines.locked = True
        self.editor.status['help'] = True
        self.editor.saved_since_edit = True
        if self.width > 80:
            self.config['page_guide'] = 72
        else:
            self.config['page_guide'] = False



import os
import threading
from copy import deepcopy
from .console import curses


class Config:
    __singleton_lock = threading.Lock()
    __singleton_instance = None

    os_name = ''
    settings = {}
    settings_path = ''
    no_bold = False
    space_char = "_"
    backup_settings = {}
    colors = {}
    debug = os.environ.get('DEBUG', 0)
    bold_text = curses.A_BOLD

    @classmethod
    def __new__(cls, *args, **kwargs):
        if not cls.__singleton_instance:
            with cls.__singleton_lock:
                if not cls.__singleton_instance:
                    cls.__singleton_instance = super().__new__(cls)
        return cls.__singleton_instance

    def __init__(self, app):
        self.app = app
        self._init_system_settings()
        self.set_default_settings()
        if not self.load_settings():
            self.save_settings()

    def _init_system_settings(self):
        self.os_name = self.get_os_name()
        user_path = os.path.expanduser('~')
        if self.os_name == 'Macintosh':
            self.settings_path = user_path + '/Library/Preferences/lexed'
        else:
            self.settings_path = user_path + '/.lexed'

    @staticmethod
    def get_os_name():
        system_info = os.uname()
        if 'Linux' in system_info:
            operating_system = 'Linux'
        elif 'Darwin' in system_info or 'Macintosh' in system_info:
            operating_system = 'Macintosh'
        else:
            operating_system = 'Other'
        return operating_system

    def set_default_settings(self, colors_only=False, skip_mono=False):
        settings = {}
        if not skip_mono:
            if self.no_bold:
                self.bold_text = 0
            else:
                self.bold_text = curses.A_BOLD
            settings['mono_normal'] = 0
            settings['mono_reverse'] = curses.A_REVERSE
            settings['mono_bold'] = self.bold_text
            settings['mono_underline'] = curses.A_UNDERLINE
            settings['mono_reverse_bold'] = curses.A_REVERSE + self.bold_text
            settings['mono_reverse_underline'] = curses.A_REVERSE + curses.A_UNDERLINE

        settings.update({key: settings[value] for key, value in {
            'color_normal': 'mono_normal',
            'color_background': 'mono_normal',
            'color_dim': 'mono_normal',
            'color_number': 'mono_normal',
            'color_warning': 'mono_reverse_bold',
            'color_message': 'mono_reverse',
            'color_reversed': 'mono_reverse',
            'color_underline': 'mono_underline',
            'color_commands': 'mono_underline',
            'color_quote_single': 'mono_bold',
            'color_quote_double': 'mono_bold',
            'color_quote_triple': 'mono_bold',
            'color_line_numbers': 'mono_normal',
            'color_line_num_reversed': 'mono_reverse',
            'color_operator': 'mono_normal',
            'color_entry': 'mono_reverse',
            'color_functions': 'mono_underline',
            'color_functions_reversed': 'mono_reverse',
            'color_commands_reversed': 'mono_reverse',
            'color_mark': 'mono_reverse_underline',
            'color_entry_command': 'mono_reverse_bold',
            'color_entry_quote': 'mono_reverse',
            'color_entry_quote_triple': 'mono_reverse',
            'color_entry_comment': 'mono_reverse',
            'color_entry_functions': 'mono_reverse_bold',
            'color_entry_class': 'mono_reverse_bold',
            'color_entry_dim': 'mono_reverse',
            'color_entry_number': 'mono_reverse',
            'color_comment': 'mono_normal',
            'color_comment_block': 'mono_reverse',
            'color_comment_separator': 'mono_reverse',
            'color_comment_leftjust': 'mono_reverse',
            'color_comment_rightjust': 'mono_reverse',
            'color_comment_centered': 'mono_reverse',
            'color_tab_odd': 'mono_normal',
            'color_tab_even': 'mono_bold',
            'color_whitespace': 'mono_reverse_underline',
            'color_class': 'mono_underline',
            'color_class_reversed': 'mono_reverse',
            'color_negative': 'mono_normal',
            'color_entry_negative': 'mono_reverse',
            'color_positive': 'mono_normal',
            'color_entry_positive': 'mono_reverse',
            'color_header': 'mono_normal',
            'color_bar': 'mono_normal',
            'color_constant': 'mono_normal',
            'color_entry_constant': 'mono_reverse',
            'color_selection': 'mono_reverse_underline',
            'color_selection_reversed': 'mono_underline',
        }.items()})
        if not colors_only:
            settings['entry_highlighting'] = True
            settings['syntax_highlighting'] = True
            settings['live_syntax'] = True
            settings['debug'] = True
            settings['collapse_functions'] = False
            settings['splitscreen'] = False
            settings['showSpaces'] = False
            settings['show_indent'] = True
            settings['inline_commands'] = True  # set to 'protected' to protect commands with protect string
            settings['protect_string'] = '::'
            settings['format_comments'] = True
            settings['display_color'] = True
            settings['default_colors'] = True
            settings['auto'] = False
            settings['page_guide'] = False
            settings['cursor_acceleration'] = True
            # The following settings can only be changed by manually editing pref file
            settings['cursor_max_horizontal_speed'] = 5
            settings['cursor_max_vertical_speed'] = 12
            settings['encrypt_warning'] = True
            settings['skip_confirmation'] = False
            settings['highlight_commands'] = True
            settings['deselect_on_copy'] = False
            settings['select_on_paste'] = False
            # settings['default_load_sort'] = 'name'
            # settings['default_load_reverse'] = False
            # settings['default_load_invisibles'] = False
            settings['key_entry_window'] = 5
            settings['key_find'] = 6
            settings['key_find_again'] = 7
            settings['key_next_bug'] = 4
            settings['key_next_marked'] = 14
            settings['key_previous_marked'] = 2
            settings['key_deselect_all'] = 1
            settings['key_page_down'] = 16
            settings['key_page_up'] = 21
            settings['key_save_as'] = 23
            settings['terminal_command'] = 'gnome-terminal -x'  # to launch in xterm, set to 'xterm -e'

        settings['default_load_sort'] = 'name'  # Had to define these here to fix bug
        settings['default_load_reverse'] = False
        settings['default_load_invisibles'] = False

        self.settings = settings

    def load_settings(self):
        result = False
        if os.path.exists(self.settings_path):
            with open(self.settings_path) as config_file:
                for line in config_file:
                    key, value = (x.strip() for x in line.strip().split('='))
                    try:
                        value = int(value)  # convert value to number if possible
                    except (ValueError, TypeError):
                        pass
                    if value == 'True':
                        value = True
                    elif value == 'False':
                        value = False
                    self.settings[key] = value
            result = True
        return result

    def save_settings(self):
        # global program_message
        editor = getattr(self.app, 'editor', None)
        if editor:
            editor.reset_line()
        settings_list = []
        for key, value in self.settings.items():
            line_of_text = f'{key} = {value}\n'
            settings_list.append(line_of_text)
        settings_list.sort()

        with open(self.settings_path, 'w') as config_file:
            config_file.writelines(settings_list)

        if editor:
            editor.program_message = ' Settings saved '

    def invert_monochrome(self):
        """Inverts monochrome display"""
        self.settings.update({
            'mono_bold': 2359296,
            'mono_normal': 262144,
            'mono_reverse': 0,
            'mono_reverse_bold': 2097152,
            'mono_reverse_underline': 131072,
            'mono_underline': 393216,
        })

    def set(self, key, value):
        self.settings[key] = value

    def get(self, key, default=None):
        return self.settings.get(key, default)

    def __getitem__(self, key, default=None):
        return self.settings.get(key, default)

    def __setitem__(self, key, value):
        self.settings[key] = value

    def copy_settings(self, reverse=False):
        """Makes a backup copy of current settings

                If optional arg reverse is set to 'True',
                settings is copied from backup"""
        # global settings, backupSettings
        if reverse:
            self.settings = deepcopy(self.backup_settings)
        else:
            self.backup_settings = deepcopy(self.settings)



import os
import sys

from .config import Config
from .console import curses
from .window import Window
from .meta import BareException


class LexEd:
    def __init__(self):
        self.clipboard = []
        self.config = Config(self)
        self.window = Window(self)
        self.editor = self.window.editor
        self.break_now = False

    def run(self):
        if self.window.height < 14 or self.window.width < 49:
            return 1, 'ERROR: lexed could not be launched, minimum terminal size of 50 x 15 required'

        # check for command line arguments
        args = sys.argv
        set_mode = False
        set_color = False
        help_flag = False
        set_protect_string = False
        read_only = False
        no_bold = False

        if '-' not in args[-1] and len(args) > 1 and not args[-1].startswith('['):  # check for file name/path
            path = args[-1]
            if '/' not in path:
                path = os.path.abspath(path)
        else:
            path = False

        if path and args and len(args) >= 2 and 's' in args[-2] and \
                args[-2].startswith('-') and not args[-2].startswith('--'):
            path = False

        if path and not os.path.exists(path):
            path = False  # New bit to stop program from crashing when path is bad
            self.editor.program_message = ' Error, file could not be loaded! '

        # check for flags
        flag_list = []
        for item in args:
            if item.startswith('-') and not item.startswith('--'):
                for letter in item:
                    if letter != '-':
                        flag_list.append(letter)
        # check for help flag
        if 'h' in flag_list or '--help' in args:
            help_flag = True
        # set mode and color
        if 't' in flag_list or '--text' in args:
            set_mode = 'text'
        elif 'p' in flag_list or '--python' in args:
            set_mode = 'python'
        if 'c' in flag_list or '--color' in args:
            set_color = 'color'
        elif 'm' in flag_list or '--mono' in args:
            set_color = 'monochrome'
        elif 'i' in flag_list or '--inverted' in args:
            set_color = 'inverted'
        if 'r' in flag_list or '--read' in args:
            read_only = True
        if 'n' in flag_list or '--nobold' in args:
            no_bold = True

        # FOR EDITING PURPOSES ONLY, REMOVE WHEN DONE!!!
        if "--source" in args and not path:
            path = args[0]  # lexed opens copy of itself for editing

        pos = 0
        if 's' in flag_list or '--string' in args:  # set protect string
            if '--string' in args:
                pos = args.index('--string')
                if path and len(args) >= 2 and 's' in args[-2] and args[-2] == '--string':
                    path = False
            else:
                if path and len(args) >= 2 and 's' in args[-2] and args[-2].startswith('-'):
                    path = False
                for i in range(0, len(args)):
                    item = args[i]
                    if item.startswith('-') and not item.startswith('--') and 's' in item:
                        pos = i

            flag_error = """
            ERROR:\n
            Argument required by string flag (maximum size 4 characters)\n
            Example: lexed -s '::' myfile.txt\n
            """
            try:
                set_protect_string = args[pos + 1]
                # next part was modified to safeguard against pathname becoming protect string
                if len(set_protect_string) < 5:
                    self.config[
                        "inline_commands"] = "protected"  # set to "protected" to protect commands with protect string
                    self.config["protect_string"] = set_protect_string
                else:
                    self.window.curses_off()
                    print(flag_error)
                    sys.exit()
            except BareException:
                self.window.curses_off()
                print(flag_error)
                sys.exit()

        if no_bold:
            self.config.set_default_settings(False, False)

        if set_mode == "text":
            self.config.settings.update({
                "entry_highlighting": False,
                "syntax_highlighting": False,
                "live_syntax": False,
                "debug": False,
                "collapse_functions": False,
                "showSpaces": False,
                "show_indent": False,
                "inline_commands": "protected",  # set to "protected" to protect commands with protect string
                "format_comments": True,
                "protect_string": set_protect_string or self.config["protect_string"],
            })
        elif set_mode == "python":
            self.config.settings.update({
                "entry_highlighting": True,
                "syntax_highlighting": True,
                "live_syntax": True,
                "debug": True,
                "show_indent": True,
                # set to "protected" to protect commands with protect string
                "inline_commands": self.config["inline_commands"] or True,
                "format_comments": True,
            })
        if set_color == "color":
            self.config.settings.update({
                "display_color": True,
                "default_colors": True,
            })
        elif set_color == "monochrome":
            self.config.settings["display_color"] = False
            self.config.set_default_settings(True, True)
        elif set_color == "inverted":
            self.config.settings["display_color"] = False
            self.config.invert_monochrome()
            self.config.set_default_settings(True, True)

        # Turn on color
        if self.config["display_color"] and self.config["default_colors"]:
            self.window.color_on(True)  # default colors
        elif self.config["display_color"]:
            self.window.color_on()  # user defined colors
        # Clear screen and draw bar
        if self.config["color_background"]:
            self.window.print_background()
        self.window.hline(1, 0, curses.ACS_HLINE, self.window.width, self.config["color_bar"])
        # Adjust page guide if necessary
        if self.config["page_guide"] > self.window.width - 7:
            self.config["page_guide"] = False
        # Load file or create new doc
        # self.editor.lines.db = {}  # create new doc (moved to fix bug)
        # self.editor.lines.add('')
        self.editor.current_num = 1
        self.editor.save_path = ''
        self.editor.saved_since_edit = True

        if help_flag:
            self.window.show_help()
            # set_mode = False
        elif path:  # load file if path exists
            # print 'splash screen' while loading
            half_height = int(self.window.height / 2)
            half_width = int(self.window.width / 2)
            self.window.clear()
            self.window.hline(1, 0, curses.ACS_HLINE, self.window.width, self.config["color_bar"])
            self.window.addstr(half_height - 1, half_width - 11, "                       ",
                               self.config["color_message"])
            self.window.addstr(half_height, half_width - 11, "    lexed  1.04    ", self.config["color_message"])
            self.window.addstr(half_height + 1, half_width - 11, "                       ",
                               self.config["color_message"])
            if read_only:  # load read only file
                self.editor.load(path, True)
            else:  # load file for editing
                self.editor.load(path)

        if not help_flag and not self.editor.lines.locked:
            self.editor.current_num = self.editor.lines.total
        self.editor.current_line = self.editor.lines.db[self.editor.current_num]

        self.editor.update_que()

        # indent_level = 0
        # c = 0

        # Begin Main Loop
        self.editor.run_editor()

        # @
        # # Version: 1.04
        # # License: Open Source
        # # Edited on 07/25/08 09:54:26 PM
        # # -------------------------------------
        # # Fixed errorTest() bug that caused crashes when loading small files
        # @



from .meta import BareException
from .const import EXECUTABLES, COMMANDS


class Line:
    """This class represents an individual line of text/code"""

    def __init__(self, lines, number, text=''):
        self.lines = lines
        self.editor = lines.editor
        self.config = lines.config
        self.number = number
        self.__text = text
        self.number_of_rows = 1
        self.row = {}
        self.calc_rows()
        self.split_rows()
        self.calc_cursor()
        self.__y = 0
        self.__x = self.end_x
        self.end_y = 0 - self.number_of_rows + 1
        self.end_x = 0
        self.length = len(text)
        self.syntax = {}
        self.listing = []
        self.executable = False
        self.collapsed = False
        self.indent_required = False
        self.error = False
        self.indentation = len(self.__text) - len(self.__text.lstrip())
        self.marked = False
        self.selected = False
        self.continue_quoting = False
        self.equal_continues = False
        self.if_continues = False
        for character in text:
            self.listing.append(character)

    def __str__(self):
        return self.__text

    def __int__(self):
        return self.lines.db

    def __add__(self, char):
        new_text = self.__text + char
        self.set_text(new_text)

    def calc_rows(self, width=None):
        """Calculates the number of rows required to print line"""
        width = width or self.editor.window.width
        number_of_lines = int(len(self.__text) / width + 1)
        self.number_of_rows = number_of_lines

    def check_executable(self):
        """Checks text to see if it contains an executable command"""
        if not self.config['inline_commands']:
            return
        if not self.config['highlight_commands']:
            return
        if not self.text:
            self.executable = False
            return
        for item in (
                ' ', '#', 'run ', 'strip ', 'prev ', 'new ', 'quit ', 'revert ', 'timestamp ', 'undo ', 'saveprefs ',
                'setcolor ', 'setcolors ', 'previous '):
            if self.text.startswith(item):
                self.executable = False
                return
        word1 = ''
        word2 = ''
        temp_list = self.text.split()
        if temp_list:
            word1 = temp_list[0]
        if len(temp_list) >= 2:
            word2 = temp_list[1]
        if self.config['inline_commands'] == 'protected':
            p_string = self.config['protect_string']
            if not word1.startswith(p_string):
                return
            word1 = word1.replace(p_string, '', 1)

        if word1 and word1 in EXECUTABLES and word1 == EXECUTABLES[EXECUTABLES.index(word1)]:
            if word1 in ('replace', 'protect') and ' with ' in self.text:
                self.executable = True

            elif len(temp_list) > 1 and word2 and word2[0] in ('=', '+', '-', '*', '/', '%', '(', '[', '{'):
                if word1 in ('save', 'saveas', 'load') and word2[0] == '/':
                    pass
                else:
                    self.executable = False

            elif word1 in (
                    'syntax', 'entry', 'live', 'formatting', 'tab', 'tabs', 'whitespace', 'show', 'hide', 'goto',
                    'color', 'help', 'debug', 'split', 'guide', 'pageguide') and self.text.count(' ') > 1:
                self.executable = False

            elif word1 in ('replace', 'protect') and \
                    self.text.count(' ') > 3 and \
                    ' with' not in self.text and \
                    '|' not in self.text:
                self.executable = False
                return

            elif word1 not in ('replace', 'protect', 'find', 'save', 'saveas', 'load', 'mark') and \
                    self.text.count(' ') - 1 > self.text.count(',') + (2 * self.text.count('-')):
                self.executable = False
                return

            elif word1 not in ('replace', 'protect', 'find', 'save', 'saveas', 'load', 'mark') and \
                    self.text.count('-') > 1:
                self.executable = False
                return

            elif word1 not in (
                    'replace', 'protect', 'find', 'save', 'saveas', 'load',
                    'mark') and '-' in self.text and ',' in self.text:
                self.executable = False
                return

            else:
                self.executable = True
        else:
            self.executable = False

    def split_rows(self, width=None):
        """This function creates a dictionary with strings of proper size"""
        width = width or self.editor.window.width
        self.row = {}
        for i in range(0, self.number_of_rows):
            self.row[i] = self.__text[(width * i):(width * i + width)]

    @staticmethod
    def num_left_of(text, position):
        """Function to compare position of number with other character"""
        flag = True
        if '[' not in text and '(' not in text and ':' not in text and ',' not in text and \
                ';' not in text and '=' not in text and '+' not in text and '-' not in text and \
                '*' not in text and '/' not in text and '%' not in text:
            return False
        
        for char in '[(:,;+-*/%=':
            if char in text:
                char_pos = text.rfind(char)
                if position <= char_pos:
                    flag = False
                    continue
                flag = True
                for i in range(char_pos, position):
                    if text[i].isalpha():
                        flag = False
        return flag

    def add_syntax(self):  # , width=None):
        """This function creates a dictionary with added syntax"""
        # width = width or self.editor.width
        self.syntax = {}
        self.continue_quoting = False
        double_quote = False
        single_quote = False
        triple_quote = False
        comment = False
        command = False
        number_char = False
        operator = False
        func = False
        separator = False
        backslash = False
        mark_on = False
        my_class = False
        # negative = False

        quote_number = 0
        mark_num = 1

        if self.number > 1 and self.continue_quoting:
            triple_quote = True
        # New section to fix multi-line quoting (with '\')
        elif self.number > 1 and \
                '_!DQT!_' in str(self.syntax.values()) and \
                '_!OFF!_' not in str(self.syntax.values()) and \
                self.text.endswith('\\'):
            double_quote = True
        elif self.number > 1 and \
                '_!SQT!_' in str(self.syntax.values()) and \
                '_!OFF!_' not in str(self.syntax.values()) and \
                self.text.endswith('\\'):
            single_quote = True

        try:
            if self.text and self.text[self.indentation] == '#':
                triple_quote = False
        except BareException:
            pass

        count = 0

        for key in range(0, len(self.row)):
            temp_list = []
            word = ''
            # phrase = ''
            old_letter = ''

            for j in range(0, len(self.row[key])):
                letter = self.row[key][j]
                try:
                    next_letter = self.row[key][j + 1]
                except BareException:
                    next_letter = ''

                word += letter

                var_found = False  # This bit used to determine when a number is part of variable
                if len(word) > 1:
                    if word[1].isalpha():
                        var_found = True
                    elif word[0].isalpha():
                        var_found = True
                elif len(word) == 1:
                    if word[0].isalpha():
                        var_found = True

                if mark_on:
                    mark_num += 1

                if self.marked and self.marked is not True and self.marked in self.row[key]:
                    try:
                        if self.row[key][j:j + len(self.marked)] == self.marked:
                            temp_list.append('_!MAR!_')
                            mark_on = True
                    except BareException:
                        pass

                if not backslash and not single_quote and not double_quote and not triple_quote:
                    try:
                        threesome = self.row[key][j:j + 3]
                        if threesome == '"""':
                            triple_quote = True
                            temp_list.append('_!TQT!_')
                            temp_list.append(letter)
                            word = ''
                            continue
                    except BareException:
                        pass

                if letter == '"':
                    quote_number += 1  # trying to fix triple quotes broken by line break
                else:
                    quote_number = 0

                if not backslash and letter == '\\':
                    if single_quote or double_quote or triple_quote:
                        backslash = True
                        temp_list.append(letter)
                        continue

                if backslash and letter in ('"', "'", '\\', 'n', 't', 'a', 'r', 'f', 'b'):  # added 'r', 'f' and 'b'
                    backslash = False
                    temp_list.append(letter)
                    continue

                if triple_quote and not temp_list:
                    temp_list.append('_!TQT!_')
                elif double_quote and not temp_list:
                    temp_list.append('_!DQT!_')
                elif single_quote and not temp_list:
                    temp_list.append('_!SQT!_')

                if number_char and letter not in '0123456789':
                    number_char = False
                    temp_list.append('_!NOF!_')

                if operator and letter not in '+-*/%=><![](){}':
                    operator = False
                    temp_list.append('_!OOF!_')

                if self.config['show_indent'] and letter == ' ' and count < self.indentation:
                    temp_list.append('_!IND!_')
                    count += 1
                elif self.config['showSpaces'] and letter == ' ' and not command and not func and not my_class:
                    temp_list.append('_!SPA!_')

                if separator and not temp_list:  # comment lines that wrap across rows
                    temp_list.append('_!SEP!_')

                if comment and not temp_list:  # comment lines that wrap across rows
                    temp_list.append('_!CMT!_')
                    temp_list.append(letter)

                elif word in ('def', 'def:') and \
                        not double_quote and \
                        not single_quote and \
                        not triple_quote and \
                        not command and \
                        not func and \
                        not my_class and \
                        not comment and \
                        not separator and \
                        next_letter in (' ', ''):
                    temp_list.insert(-len(word) + 1, '_!FUN!_')
                    func = True
                    word = ''
                    temp_list.append(letter)

                elif word in ('class', 'class:') and \
                        not double_quote and \
                        not single_quote and \
                        not triple_quote and \
                        not command and \
                        not func and \
                        not my_class and \
                        not comment and \
                        not separator and \
                        next_letter in (' ', ''):
                    temp_list.insert(-len(word) + 1, '_!CLA!_')
                    my_class = True
                    word = ''
                    temp_list.append(letter)

                elif word.endswith("False") and \
                        not word[word.rfind("False") - 1:].isalnum() and \
                        not double_quote and not single_quote and \
                        not triple_quote and not comment and \
                        not separator and not func and \
                        next_letter in (' ', '', ')', ',', ';', ']', '}', ':'):
                    temp_list.insert(-len('False') + 1, '_!NEG!_')
                    # negative = True
                    word = ''
                    temp_list.append(letter)
                    temp_list.append('_!NEO!_')

                elif word[word.rfind('(') + 1:] in ('not', 'False', 'False,', 'False)', ', False)', 'False:') and \
                        not double_quote and not single_quote and \
                        not triple_quote and not comment and \
                        not separator and not func and \
                        next_letter in (' ', '', ')', ',', ';', ']', '}', ':'):
                    if '(' in word:
                        temp_list.insert(-len(word[word.rfind('(') + 1:]) + 1, '_!NEG!_')
                    elif ':' in word:
                        temp_list.insert(-len(word[word.find(':') + 1:]) + 1, '_!NEG!_')
                    else:
                        temp_list.insert(-len(word) + 1, '_!NEG!_')
                    # negative = True
                    word = ''
                    temp_list.append(letter)
                    temp_list.append('_!NEO!_')

                elif word.endswith('True') and not word[word.rfind('True') - 1:].isalnum() and \
                        not double_quote and not single_quote and not triple_quote and \
                        not comment and not separator and not func and not next_letter.isalnum():
                    temp_list.insert(-len('True') + 1, '_!POS!_')
                    # positive = True
                    word = ''
                    temp_list.append(letter)
                    temp_list.append('_!POO!_')

                elif word[word.rfind('(') + 1:] in ('True', 'True,', 'True)', ',True)', 'True:') and \
                        not double_quote and not single_quote and not triple_quote and \
                        not comment and not separator and not func and \
                        next_letter in (' ', '', ')', ',', ';', ']', '}', ':'):
                    if '(' in word:
                        temp_list.insert(-len(word[word.rfind('(') + 1:]) + 1, '_!POS!_')
                    elif ':' in word:
                        temp_list.insert(-len(word[word.find(':') + 1:]) + 1, '_!POS!_')
                    else:
                        temp_list.insert(-len(word) + 1, '_!POS!_')
                    # positive = True
                    word = ''
                    temp_list.append(letter)
                    temp_list.append('_!POO!_')

                # New bit for CONSTANT variables
                elif not next_letter.isalnum() and next_letter != '_' and \
                        not single_quote and not double_quote and \
                        not triple_quote and not comment and not separator and \
                        word[word.rfind('(') + 1:].isupper() and "'" not in word and '"' not in word and \
                        len(word[word.rfind('(') + 1:]) > 2 and word[word.rfind('(') + 1].isalpha() and \
                        '+' not in word and '-' not in word and '*' not in word and '/' not in word:
                    if '(' in word:
                        temp_list.insert(-len(word[word.rfind('(') + 1:]) + 1, '_!CON!_')
                    else:
                        temp_list.insert(-len(word) + 1, '_!CON!_')
                    word = ''
                    temp_list.append(letter)
                    temp_list.append('_!COO!_')

                # New bit for CONSTANT variables (handling ':')
                elif not next_letter.isalnum() and next_letter != '_' and not single_quote and \
                        not double_quote and not triple_quote and not comment and not separator and \
                        word[word.rfind(':') + 1:].isupper() and "'" not in word and '"' not in word and \
                        len(word[word.rfind(':') + 1:]) > 2 and word[word.rfind(':') + 1].isalpha() and \
                        '+' not in word and '-' not in word and '*' not in word and '/' not in word:
                    if ':' in word:
                        temp_list.insert(-len(word[word.rfind(':') + 1:]) + 1, '_!CON!_')
                    else:
                        temp_list.insert(-len(word) + 1, '_!CON!_')
                    word = ''
                    temp_list.append(letter)
                    temp_list.append('_!COO!_')

                # New bit for CONSTANT variables (handling '[')
                elif not next_letter.isalnum() and next_letter != '_' and not single_quote and \
                        not double_quote and not triple_quote and not comment and not separator and \
                        word[word.rfind('[') + 1:].isupper() and "'" not in word and '"' not in word and \
                        len(word[word.rfind('[') + 1:]) > 2 and word[word.rfind('[') + 1].isalpha() and \
                        '+' not in word and '-' not in word and '*' not in word and '/' not in word:
                    if '[' in word:
                        temp_list.insert(-len(word[word.rfind('[') + 1:]) + 1, '_!CON!_')
                    else:
                        temp_list.insert(-len(word) + 1, '_!CON!_')
                    word = ''
                    temp_list.append(letter)
                    temp_list.append('_!COO!_')

                elif word in COMMANDS and not double_quote and not single_quote and not triple_quote and \
                        not command and not func and not my_class and not comment and not separator and \
                        next_letter in (' ', '', ',', ';'):
                    temp_list.insert(-len(word) + 1, '_!CMD!_')
                    command = True
                    word = ''  # changed from one space
                    temp_list.append(letter)

                elif func and letter in ' :=<>.()[]abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ':
                    func = False
                    temp_list.append('_!FOF!_')
                    if letter == ' ' and self.config['showSpaces']:
                        temp_list.append('_!SPA!_')
                    temp_list.append(letter)

                elif my_class and letter in ' :=<>.()[]abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ':
                    my_class = False
                    temp_list.append('_!FOF!_')
                    if letter == ' ' and self.config['showSpaces']:
                        temp_list.append('_!SPA!_')
                    temp_list.append(letter)

                elif command and letter in ' :=<>.()[]abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ':
                    command = False
                    temp_list.append('_!NOC!_')
                    if letter == ' ' and self.config['showSpaces']:
                        temp_list.append('_!SPA!_')
                    temp_list.append(letter)

                # New part to check for triple_quotes
                elif letter == '"' and '"""' in word and not backslash and not single_quote:
                    if triple_quote and '\\' in word:
                        try:
                            threesome = self.row[key][j - 2:j + 1]
                            four_ago = self.row[key][j - 3]
                            if threesome == '"""' and four_ago != '\\':
                                triple_quote = False
                                temp_list.append(letter)
                                temp_list.append('_!OFF!_')
                                word = ''
                                self.continue_quoting = False
                            else:
                                temp_list.append(letter)
                        except BareException:
                            pass

                    elif triple_quote:
                        triple_quote = False
                        temp_list.append(letter)
                        temp_list.append('_!OFF!_')
                        word = ''
                        self.continue_quoting = False
                    else:
                        triple_quote = True
                        temp_list.append('_!TQT!_')
                        temp_list.append(letter)
                        word = ''
                elif letter == '"' and not backslash and triple_quote and quote_number == 3:
                    # new bit to fix triple quotes broken by line break
                    triple_quote = False
                    temp_list.append(letter)
                    temp_list.append('_!OFF!_')
                    word = ''
                    self.continue_quoting = False

                elif letter == '"' and not backslash and \
                        not triple_quote and not double_quote and \
                        not single_quote and not comment and not separator:
                    double_quote = True
                    temp_list.append('_!DQT!_')
                    temp_list.append(letter)
                elif letter == "'" and not backslash and \
                        not triple_quote and not double_quote and \
                        not single_quote and not comment and not separator:
                    single_quote = True
                    temp_list.append('_!SQT!_')
                    temp_list.append(letter)
                elif letter == '"' and double_quote and not backslash:
                    double_quote = False
                    temp_list.append(letter)
                    temp_list.append('_!OFF!_')
                elif letter == "'" and single_quote and not backslash:
                    single_quote = False
                    temp_list.append(letter)
                    temp_list.append('_!OFF!_')

                elif letter == '#' and not double_quote and not single_quote and not triple_quote and not separator:
                    if next_letter == '#':
                        separator = True
                        temp_list.append('_!SEP!_')
                    else:
                        comment = True
                        temp_list.append('_!CMT!_')
                    temp_list.append(letter)

                # new number check
                elif letter.isdigit() and not var_found and not number_char and not double_quote and \
                        not single_quote and not triple_quote and not comment and not separator:
                    number_char = True
                    temp_list.append('_!NUM!_')
                    temp_list.append(letter)

                elif letter.isdigit() and not number_char and self.num_left_of(word, len(word)) and \
                        not old_letter.isalpha() and not double_quote and not single_quote and \
                        not triple_quote and not comment and not separator:
                    number_char = True
                    temp_list.append('_!NUM!_')
                    temp_list.append(letter)

                elif letter in '01234567890' and not number_char and j == 0 and not double_quote and \
                        not single_quote and not triple_quote and not comment and not separator:
                    number_char = True
                    temp_list.append('_!NUM!_')
                    temp_list.append(letter)

                elif letter in '+-*/%=><![](){}' and not number_char and not double_quote and \
                        not single_quote and not triple_quote and not comment and not separator:
                    operator = True
                    temp_list.append('_!OPE!_')
                    temp_list.append(letter)

                else:
                    temp_list.append(letter)
                if letter == ' ' or letter == '    ':
                    word = ''

                if mark_on and mark_num == len(self.marked):
                    temp_list.append('_!MOF!_')
                    mark_on = False
                    mark_num = 1

                old_letter = letter

            self.syntax[key] = temp_list
        if triple_quote:
            self.continue_quoting = True

    def set_text(self, text):
        """Changes line text and updates other attributes"""
        if self.lines.locked:
            return  # don't allow editing if locked
        # old_len = len(self.__text)
        self.__text = text
        self.calc_rows()
        self.split_rows()
        self.calc_cursor()
        self.length = len(text)
        self.listing = []
        for character in text:
            self.listing.append(character)
        self.end_y = 0 - self.number_of_rows + 1
        self.indentation = len(self.__text) - len(self.__text.lstrip())
        self.check_executable()
        # Add syntax if line other than current line is edited (fix acceleration bug)
        if self.config['syntax_highlighting'] and self.number != self.editor.current_num:
            self.add_syntax()

    def get_text(self):
        return self.__text

    def setx(self, num):
        self.__x = num
        if self.__x < 6 and self.__y >= self.end_y:
            self.__x = self.editor.window.width - 1
            self.__y -= 1
            if self.__y < self.end_y:
                self.__y = 0
                self.__x = self.end_x
        elif self.__x < 6 and self.__y != 0:
            self.__x = self.editor.window.width - 1
        elif self.__x < 6:
            self.__x = self.end_x
        elif self.__x > self.editor.window.width - 1 and self.__y < 0:
            self.__x = 6
            self.__y += 1
        elif self.__x > self.end_x and self.__y == 0:
            self.__y = self.end_y
            self.__x = 6

    def sety(self, num):
        self.__y = num
        if self.__y < self.end_y:
            self.__y = 0
        elif self.__y > 0:
            self.__y = 0

    def getx(self):
        return self.__x

    def gety(self):
        return self.__y

    def swap(self, obj):
        temp_text = self.__text
        self.set_text(obj.__text)
        obj.set_text(temp_text)

    def calc_cursor(self):
        self.end_x = len(self.row[(self.number_of_rows - 1)]) + 6

    text = property(get_text, set_text)
    x = property(getx, setx)
    y = property(gety, sety)


class Lines:
    def __init__(self, editor):
        self.editor = editor
        self.config = editor.config
        self.window = editor.window
        self.number = 0
        self.total = 0
        self.locked = False
        self.db = {}

    def add(self, text=''):
        self.total = len(self.db) + 1
        self.db[self.total] = Line(self, self.total, text)
        return self.db[self.total]

    def renum(self):
        """Renumber lines"""
        current_num = 1
        for key, value in self.db.items():
            # temp = value
            self.db[current_num] = value
            self.db[current_num].number = current_num  # new bit

            if value.number > current_num:
                del self.db[key]
                self.db[1].recalc_total()
            current_num += 1

    def recalc_total(self):
        self.total = len(self.db)

    @staticmethod
    def end_colon(text):
        """Finds end colon in text even when followed by comment
                Returns 'True' if colon found, 'False' otherwise"""
        if text and text[-1] == ':':
            return True
        elif ':' in text:
            temp_list = text.split(':')
            prev = temp_list[-2]
            last = temp_list[-1]
            if last.strip().startswith('#') and not prev.rstrip().endswith('#'):
                return True
            else:
                return False
        else:
            return False

    def collapse_these(self, num):
        """Helper function to process redundant text checks in collapse function"""
        item = self.db[num]
        if item.text.endswith(':') or self.end_colon(item.text):
            self.collapse_line(num)

    def collapse(self, text):
        """Function that attempts to collapse lines"""
        # global program_message
        self.editor.program_message = ' Collapsed lines '
        selection = False
        func_total = 0
        if text == 'collapse':
            selection, item_count = self.editor.get_selected()
            if selection:
                text = 'collapse %s' % selection
        temp_text = text
        self.editor.reset_line()

        num = 0
        item_found = ''
        search_string = ''
        try:
            # Search for function or class
            if len(temp_text) > 9 and temp_text[9].isalpha() and \
                    self.editor.get_args(temp_text) not in ('function', 'functions', 'class', 'classes'):
                search_string = temp_text[9:]
                find_function = 'def ' + search_string + '('
                find_class = 'class ' + search_string + '('
                for i in range(1, len(self.db) + 1):
                    item = self.db[i]
                    if item.text.strip().startswith(find_function) or item.text.strip().startswith(find_class):
                        if item.text.strip().startswith('def'):
                            item_found = 'function'
                        elif item.text.strip().startswith('class'):
                            item_found = 'class'
                        temp_text = 'collapse %i' % i
                        func_total += 1
                        break
                if not func_total:
                    self.editor.program_message = ' Specified function/class not found! '
                    return

            if self.editor.get_args(temp_text) in ('function', 'functions', 'class', 'classes'):
                # settings['collapse_functions'] = True
                self.collapse_functions()
            elif 'tab' in temp_text or 'indent' in temp_text:
                indent_num = int(self.editor.get_args(temp_text)[1])
                self.editor.program_message = ' Collapsed lines with indent of %i ' % indent_num
                for i in range(1, len(self.db) + 1):
                    self.editor.status_message('Processing: ', (100 / (self.total * 1.0 / i)))
                    # item = self.db[i]
                    try:
                        _next = self.db[i + 1]
                    except BareException:
                        return
                    if _next.indentation >= indent_num * 4:
                        self.collapse_these(i)
            elif '-' in temp_text:
                temp_list = self.editor.get_args(temp_text, ' ', '-')
                start = max(1, int(temp_list[0]))
                end = min(self.total, int(temp_list[1]))
                self.editor.program_message = ' Collapsed lines between %i and %i ' % (start, end)
                for i in range(start, end + 1):
                    self.editor.status_message('Processing: ', (100 / ((end - start) * 1.0 / i)))
                    self.collapse_these(i)
            elif ',' in temp_text:
                temp_list = self.editor.get_args(temp_text, ' ', ',')
                self.editor.program_message = ' Collapsed %i lines ' % len(temp_list)
                for i in range(0, len(temp_list)):
                    num = int(temp_list[i])
                    self.collapse_these(num)
            else:
                arg_list = self.editor.get_args(temp_text)
                if 'str' in str(type(arg_list)):
                    num = int(arg_list)
                    self.collapse_these(num)
                else:
                    for i in range(0, len(arg_list)):
                        num = int(arg_list[i])
                        self.collapse_these(num)
                self.editor.program_message = ' Collapsed line number %i ' % num
            if selection:
                self.editor.program_message = ' Selection collapsed '
            elif func_total:
                self.editor.program_message = " Collapsed %s '%s' " % (item_found, search_string)
        except BareException:
            self.editor.program_message = ' Error, collapse failed! '

    def collapse_functions(self):
        """Collapses all function and class definitions"""
        # global program_message
        for i in range(2, len(self.db) + 1):
            self.editor.status_message('Processing: ', (100 / (self.total * 1.0 / i)))
            prev_item = self.db[i - 1]
            item = self.db[i]
            if not prev_item.collapsed:
                if prev_item.text[prev_item.indentation:(prev_item.indentation + 4)] == 'def ':
                    item.collapsed = True
                    item.indent_required = prev_item.indentation
                elif prev_item.text[prev_item.indentation:(prev_item.indentation + 6)] == 'class ':
                    item.collapsed = True
                    item.indent_required = prev_item.indentation
            elif prev_item.collapsed:
                if not item.text:
                    item.collapsed = True
                    item.indent_required = prev_item.indent_required
                elif item.indentation <= prev_item.indent_required:
                    item.collapsed = False
                else:
                    item.collapsed = True
                    item.indent_required = prev_item.indent_required
            self.editor.program_message = ' Functions & classes collapsed '

    def collapse_line(self, number):
        """Attempt to collapse line number"""
        for i in range(number + 1, len(self.db)):
            item = self.db[i]
            if item.indentation > self.db[number].indentation or not item.text:
                if item.collapsed:
                    item.indent_required = min(item.indent_required, self.db[number].indentation)
                else:
                    item.indent_required = self.db[number].indentation
                item.collapsed = True
            else:
                return

    def expand(self, text):
        """Processes 'expand' command"""
        # global program_message
        selection = False
        func_total = 0
        if text == 'expand':
            selection, item_count = self.editor.get_selected()
            if selection:
                text = 'expand %s' % selection

        self.editor.reset_line()
        num = 0
        item_found = ''
        search_string = ''
        try:
            if len(text) > 7 and text[7].isalpha():  # Search for function or class
                search_string = text[7:]
                find_function = 'def ' + search_string + '('
                find_class = 'class ' + search_string + '('
                for i in range(1, len(self.db) + 1):
                    item = self.db[i]
                    if item.text.strip().startswith(find_function) or item.text.strip().startswith(find_class):
                        if item.text.strip().startswith('def'):
                            item_found = 'function'
                        elif item.text.strip().startswith('class'):
                            item_found = 'class'
                        text = 'collapse %i' % i
                        func_total += 1
                        break
                if not func_total:
                    self.editor.program_message = ' Specified function/class not found! '
                    return

            if 'expand all' in text:
                self.expand_all()
                self.editor.program_message = ' Expanded all lines '
            elif "-" in text:  # expand range of lines
                temp_list = self.editor.get_args(text, ' ', '-')
                start = max(1, int(temp_list[0]))
                end = min(self.total, int(temp_list[1]))
                self.editor.program_message = ' Expanded lines %i to %i ' % (start, end)
                for i in range(start, end + 1):
                    self.expand_line(i)
            elif ',' in text:  # expand list of lines
                temp_list = self.editor.get_args(text, ' ', ',')
                self.editor.program_message = ' Expanded %i lines ' % len(temp_list)
                for i in range(0, len(temp_list)):
                    num = int(temp_list[i])
                    self.expand_line(num)
            elif 'functions' in text:  # expand functions
                self.editor.program_message = ' Expanded functions & classes '
                for i in range(1, len(self.db) + 1):
                    item = self.db[i]
                    if item.text[item.indentation:(item.indentation + 4)] == 'def ':
                        self.editor.expand_line(i)
                    elif item.text[item.indentation:(item.indentation + 6)] == 'class ':
                        self.editor.expand_line(i)
            else:  # expand line number
                arg_list = self.editor.get_args(text)
                if 'str' in str(type(arg_list)):
                    num = int(arg_list)
                    self.expand_line(num)
                else:
                    for i in range(0, len(arg_list)):
                        num = int(arg_list[i])
                        self.expand_line(num)
                self.editor.program_message = ' Expanded line number %i ' % num
            if selection:
                self.editor.program_message = ' Selection expanded '
            elif func_total:
                self.editor.program_message = " Expanded %s '%s' " % (item_found, search_string)
        except BareException:
            self.editor.program_message = ' Error, expand failed! '

    def expand_line(self, number):
        """Attempt to expand line number"""
        if self.db[number].collapsed:
            return
        for i in range(number + 1, len(self.db) + 1):
            item = self.db[i]
            if item.indentation > self.db[number].indentation or not item.text:
                item.collapsed = False
            elif item.indentation <= self.db[number].indentation:
                return

    def expand_all(self):
        """Expand all lines"""
        # global program_message
        self.editor.program_message = " Expanded all lines "
        for i in range(1, len(self.db) + 1):
            self.db[i].collapsed = False
